/*
 *  -- micMAGMA (version 0.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2012
 *
       @generated c Tue Mar 12 12:57:59 2013
 *
 **/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

// Flops formula
#define PRECISION_s
#if defined(PRECISION_z) || defined(PRECISION_c)
#define FLOPS(m, n) ( 6. * FMULS_GETRF(m, n) + 2. * FADDS_GETRF(m, n) )
#else
#define FLOPS(m, n) (      FMULS_GETRF(m, n) +      FADDS_GETRF(m, n) )
#endif

float get_LU_error(magma_int_t M, magma_int_t N, 
		   magmaFloatComplex *A,  magma_int_t lda, 
		   magmaFloatComplex *LU, magma_int_t *IPIV)
{
    magma_int_t min_mn = min(M,N);
    magma_int_t ione   = 1;
    magma_int_t i, j;
    magmaFloatComplex alpha = MAGMA_S_ONE;
    magmaFloatComplex beta  = MAGMA_S_ZERO;
    magmaFloatComplex *L, *U;
    float work[1], matnorm, residual;
                       
    TESTING_MALLOC( L, magmaFloatComplex, M*min_mn);
    TESTING_MALLOC( U, magmaFloatComplex, min_mn*N);
    memset( L, 0, M*min_mn*sizeof(magmaFloatComplex) );
    memset( U, 0, min_mn*N*sizeof(magmaFloatComplex) );

    lapackf77_claswp( &N, A, &lda, &ione, &min_mn, IPIV, &ione);
    lapackf77_clacpy( MagmaLowerStr, &M, &min_mn, LU, &lda, L, &M      );
    lapackf77_clacpy( MagmaUpperStr, &min_mn, &N, LU, &lda, U, &min_mn );

    for(j=0; j<min_mn; j++)
        L[j+j*M] = MAGMA_S_MAKE( 1., 0. );
    
    matnorm = lapackf77_clange("f", &M, &N, A, &lda, work);

    blasf77_cgemm("N", "N", &M, &N, &min_mn,
                  &alpha, L, &M, U, &min_mn, &beta, LU, &lda);

    for( j = 0; j < N; j++ ) {
        for( i = 0; i < M; i++ ) {
            LU[i+j*lda] = MAGMA_S_SUB( LU[i+j*lda], A[i+j*lda] );
	}
    }
    residual = lapackf77_clange("f", &M, &N, LU, &lda, work);

    TESTING_FREE(L);
    TESTING_FREE(U);

    return residual / (matnorm * N);
}

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing cgetrf
*/
int main( int argc, char** argv)
{
    magma_timestr_t       start, end;
    real_Double_t flops, mic_perf, cpu_perf, error;
    magmaFloatComplex *h_A, *h_R;
    magmaFloatComplex *d_A;
    magma_int_t     *ipiv, *new_ipiv;

    /* Matrix size */
    magma_int_t M = 0, N = 0, n2, lda, ldda;
    //magma_int_t size[10] = {960,1920,3072,4032,4992,5952,7104,8064,9024,9984};
    magma_int_t size[24] = {960, 1920,3072,4032,4992,5952,7104,8064,9024,10560, 12596, 14560, 16064, 18928, 20160, 21256, 22596, 23560, 24364, 25500, 26694, 27892, 27892, 27892};
    magma_int_t i, info, min_mn, nb, ret;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};

    if (argc != 1){
        for(i = 1; i<argc; i++){
            if (strcmp("-N", argv[i])==0)
                N = atoi(argv[++i]);
            else if (strcmp("-M", argv[i])==0)
                M = atoi(argv[++i]);
        }
        if (M>0 && N>0)
            printf("  testing_sgetrf -M %d -N %d\n\n", M, N);
        else
            {
                printf("\nUsage: \n");
                printf("  testing_sgetrf -M %d -N %d\n\n", 1024, 1024);
                exit(1);
            }
    }
    else {
        printf("\nUsage: \n");
        printf("  testing_sgetrf_mic -M %d -N %d\n\n", 1024, 1024);
        M = N = size[23];
    }

    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;

    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    }

    ldda   = ((M + 31) / 32) * 32;
    lda = ldda   = (ldda % 256 == 0) ? ldda + 32 : ldda;
    n2     = lda * N;
    min_mn = min(M, N);
    nb     = magma_get_cgetrf_nb(min_mn);

    /* Allocate host memory for the matrix */
    TESTING_MALLOC_HOST(    ipiv, magma_int_t, min_mn     );
    TESTING_MALLOC_HOST(    new_ipiv, magma_int_t, min_mn     );
    TESTING_MALLOC_HOST(    h_A, magmaFloatComplex, n2     );
    TESTING_MALLOC_HOST( h_R, magmaFloatComplex, n2     );
    TESTING_MALLOC_DEV(  d_A, magmaFloatComplex, ldda*N );

    printf("\n\n");
    printf("  M     N   CPU GFlop/s    MIC GFlop/s   ||PA-LU||/(||A||*N)\n");
    printf("============================================================\n");
	fflush(0);
	for(i = 0; i < 24; i++){
		if (argc == 1){
			M = N = size[i];
		}
		min_mn= min(M, N);
		lda   = ((M + 31) / 32) * 32;
		lda   = (lda % 256 == 0) ? lda + 32 : lda;
		n2    = lda*N;
		ldda   = ((M + 31) / 32) * 32;
		ldda   = (ldda % 256 == 0) ? ldda + 32 : ldda;
		flops = FLOPS( (float)M, (float)N ) / 1000000;

		/* Initialize the matrix */
		lapackf77_clarnv( &ione, ISEED, &n2, h_A );
		lapackf77_clacpy( MagmaUpperLowerStr, &M, &N, h_A, &lda, h_R, &lda );

		/* =====================================================================
		   Performs operation using LAPACK
		   =================================================================== */
		//start = get_current_time();
		//lapackf77_cgetrf(&M, &N, h_A, &lda, ipiv, &info);
		//end = get_current_time();

		//if (info < 0)
		//	printf("Argument %d of sgetrf had an illegal value.\n", -info);

		//cpu_perf = flops / GetTimerValue(start, end);

		/* ====================================================================
		   Performs operation using MAGMA
		   =================================================================== */
		magma_csetmatrix( M, N, h_R, 0, lda, d_A, 0, ldda, queue );
		ret = magma_cgetrf_mic( M, N, d_A, 0, ldda, new_ipiv, &info, queue);
		
		magma_csetmatrix( M, N, h_R, 0, lda, d_A, 0, ldda, queue );
		start = get_current_time();
		ret = magma_cgetrf_mic( M, N, d_A, 0, ldda, new_ipiv, &info, queue);
		end = get_current_time();
		if (info < 0)
		    printf("Argument %d of cgetrf had an illegal value.\n", -info);
		if (ret != MAGMA_SUCCESS)
			printf("magma_cgetrf_mic returned with error code %d\n", ret);

		mic_perf = flops / GetTimerValue(start, end);

		/* =====================================================================
		   Check the factorization
		   =================================================================== */
		//magma_cgetmatrix( M, N, d_A, 0, ldda, h_A, 0, lda, queue );
		//error = get_LU_error(M, N, h_R, lda, h_A, new_ipiv);

		printf("%5d %5d  %6.2f         %6.2f         %e\n",
				M, N, cpu_perf, mic_perf, error);

		fflush(0);
		if (argc != 1)
			break;
	}

	/* Memory clean up */
	TESTING_FREE_DEV( d_A );
	TESTING_FREE_HOST( ipiv );
	TESTING_FREE_HOST( new_ipiv );
	TESTING_FREE_HOST( h_A );
	TESTING_FREE_HOST( h_R );

	/* Shutdown */
	magma_queue_destroy( queue );
	magma_finalize();
}
