/*
 *  -- MAGMA (version 1.2.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     November 2012
 *
 * @generated c Tue Mar 12 12:57:59 2013
 *
 **/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

// Flops formula
#define PRECISION_c
#if defined(PRECISION_z) || defined(PRECISION_c)
#define FLOPS(m, n) ( 6. * FMULS_GETRF(m, n) + 2. * FADDS_GETRF(m, n) )
#else
#define FLOPS(m, n) (      FMULS_GETRF(m, n) +      FADDS_GETRF(m, n) )
#endif

float get_LU_error(magma_int_t M, magma_int_t N, 
                    magmaFloatComplex *A,  magma_int_t lda, 
                    magmaFloatComplex *LU, magma_int_t *IPIV)
{
    magma_int_t min_mn = min(M,N);
    magma_int_t ione   = 1;
    magma_int_t i, j;
    magmaFloatComplex alpha = MAGMA_C_ONE;
    magmaFloatComplex beta  = MAGMA_C_ZERO;
    magmaFloatComplex *L, *U;
    float work[1], matnorm, residual;
                       
    TESTING_MALLOC( L, magmaFloatComplex, M*min_mn);
    TESTING_MALLOC( U, magmaFloatComplex, min_mn*N);
    memset( L, 0, M*min_mn*sizeof(magmaFloatComplex) );
    memset( U, 0, min_mn*N*sizeof(magmaFloatComplex) );

    lapackf77_claswp( &N, A, &lda, &ione, &min_mn, IPIV, &ione);
    lapackf77_clacpy( MagmaLowerStr, &M, &min_mn, LU, &lda, L, &M      );
    lapackf77_clacpy( MagmaUpperStr, &min_mn, &N, LU, &lda, U, &min_mn );

    for(j=0; j<min_mn; j++)
        L[j+j*M] = MAGMA_C_MAKE( 1., 0. );
    
    matnorm = lapackf77_clange("f", &M, &N, A, &lda, work);

    blasf77_cgemm("N", "N", &M, &N, &min_mn,
                  &alpha, L, &M, U, &min_mn, &beta, LU, &lda);

    for( j = 0; j < N; j++ ) {
        for( i = 0; i < M; i++ ) {
            LU[i+j*lda] = MAGMA_C_SUB( LU[i+j*lda], A[i+j*lda] );
        }
    }
    residual = lapackf77_clange("f", &M, &N, LU, &lda, work);

    TESTING_FREE(L);
    TESTING_FREE(U);

    return residual / (matnorm * N);
}

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing cgetrf
   */
int main( int argc, char** argv)
{
	magma_init();

	#define TESTING_CGETRF_MMIC_CHECK

	magma_timestr_t  start, end;
	real_Double_t           flops, mic_perf, cpu_perf, error;
	magmaFloatComplex *h_A, *h_R;
	magmaFloatComplex *d_lA[4] = {NULL, NULL, NULL, NULL};
	magma_int_t     *ipiv, *new_ipiv;

	/* Matrix size */
	magma_int_t M = 0, N = 0, flag = 0, n2, lda, ldda, num_mics, num_mics0 = 1, n_local;
	magma_int_t size[37] = {960, 1920,3072,4032,4992,5952,7104,8064,9024,10560, 12596, 14560, 16064, 18928, 20160, 21256, 22596, 23560, 24364, 25500, 26694, 27892, 28776, 29680, 30892, 31784, 32899, 33900, 35170, 36668, 37800, 38896, 40358, 41478, 42680, 43996, 44720};	
	magma_int_t n_size = 36;

	magma_int_t i, k, info, min_mn, nb0, nb, nk, maxn, ret;
	magma_int_t ione     = 1;
	magma_int_t ISEED[4] = {0,0,0,1};

	magma_queue_t queue;
	magma_device_t device;
	int num = 0;
	magma_err_t err;

	err = magma_get_devices( &device, 1, &num );
	if ( err != MAGMA_SUCCESS or num < 1 ) {
		fprintf( stderr, "magma_get_devices failed: %d\n", err );
		exit(-1);
	}
	err = magma_queue_create( device, &queue );
	if ( err != MAGMA_SUCCESS ) {
		fprintf( stderr, "magma_queue_create failed: %d\n", err );
		exit(-1);
	}


	if (argc != 1){
		for(i = 1; i<argc; i++){
			if (strcmp("-N", argv[i])==0) {
				N = atoi(argv[++i]);
				flag = 1;
			} else if (strcmp("-M", argv[i])==0) {
				M = atoi(argv[++i]);
				flag = 1;
			} else if (strcmp("-NMIC", argv[i])==0)
				num_mics0 = atoi(argv[++i]);
		}
	}

	if( flag != 0 ) {
		if (M>0 && N>0 && num_mics0>0)
			printf("  testing_cgetrf_mmic -M %d -N %d -NMIC %d\n\n", (int) M, (int) N, (int) num_mics0);
		else {
			printf("\nError: \n");
			printf("  (m, n, num_mics)=(%d, %d, %d) must be positive.\n\n", (int) M, (int) N, (int) num_mics0);
			exit(1);
		}
	} else {
		M = N = size[n_size-1];
		printf("\nDefault: \n");
		printf("  testing_cgetrf_mmic -M %d -N %d -NMIC %d\n\n", (int) M, (int) N, (int) num_mics0);
	}

	ldda   = ((M + 31) / 32) * 32;
	ldda   = (ldda % 256 == 0) ? (ldda + 32) : ldda; 
	n2     = ldda * N;
	min_mn = min(M, N);
	nb     = magma_get_cgetrf_nb(M);
	num_mics = num_mics0;

	/* Allocate host memory for the matrix */
	TESTING_MALLOC_HOST(ipiv, magma_int_t, min_mn);
	TESTING_MALLOC_HOST(new_ipiv, magma_int_t, min_mn);
	TESTING_MALLOC_HOST(    h_A, magmaFloatComplex, n2     );
	TESTING_MALLOC_HOST(    h_R, magmaFloatComplex, n2     );


	/* allocate device memory, assuming fixed nb and num_mics */
	for(i = 0; i < num_mics; i++){
		n_local = ((N / nb) / num_mics) * nb;
		if (i < ((N / nb) % num_mics))
			n_local += nb;
		else if (i == (N / nb) % num_mics)
			n_local += N % nb;
		micSetDevice(i + 1);

		if(n_local > 0) {
			TESTING_MALLOC_DEV( d_lA[i], magmaFloatComplex, ldda * n_local );
		} else {
			d_lA[i] = 0;
  		}
	}

	micSetDevice(1);
	nb0 = nb;

	printf("  M     N   CPU GFlop/s    MIC GFlop/s   ||PA-LU||/(||A||*N)\n");
	printf("============================================================\n");
	for(i = 0; i < n_size; i++){
		if (flag == 0){
			M = N = size[i];
		}
		min_mn= min(M, N);
		lda   = M;
		n2    = lda * N;
		ldda  = ((M + 31) / 32) * 32;
		ldda   = (ldda % 256 == 0) ? (ldda + 32) : ldda; 
		flops = FLOPS( (float)M, (float)N ) / 1000000;

		/* Initialize the matrix */
		lapackf77_clarnv( &ione, ISEED, &n2, h_A );
		lapackf77_clacpy( MagmaUpperLowerStr, &M, &N, h_A, &lda, h_R, &lda );

		/* =====================================================================
		   Performs operation using LAPACK
		   =================================================================== */
		//start = get_current_time();
		//lapackf77_cgetrf(&M, &N, h_A, &lda, ipiv, &info);
		//end = get_current_time();
		//if (info < 0) {
		//	printf("Argument %d of cgetrf had an illegal value.\n", (int) -info);
		//	break;
		//} else if (info != 0 ) {
		//	printf("cgetrf returned info=%d.\n", (int) info);
		//	break;
		//}
		//cpu_perf = flops / GetTimerValue(start, end);
		lapackf77_clacpy( MagmaUpperLowerStr, &M, &N, h_R, &lda, h_A, &lda );

		/* ====================================================================
		   Performs operation using MAGMA
		   =================================================================== */
		/* == distributing the matrix == */
		nb = magma_get_cgetrf_nb(M);
		#ifdef TESTING_CGETRF_MMIC_CHECK
		if( nb != nb0 ) {
			printf( " different nb used for memory-allocation (%d vs. %d)\n", (int) nb, (int) nb0 );
		}
		#endif
		if( num_mics0 > N / nb ) {
			num_mics = N / nb;
			if( N % nb != 0 ) num_mics ++;
			printf( " * too many MICs for the matrix size, using %d MICs\n", (int) num_mics );
		} else {
			num_mics = num_mics0;
		}


		for(size_t j = 0; j < N; j += nb){
			k = (j / nb) % num_mics;
			micSetDevice(k + 1);
			nk = min(nb, N - j);
			magma_csetmatrix( M, nk,
					h_R + j * lda, 0,                      lda,
					d_lA[k] + j / (nb * num_mics) * nb * ldda, 0, ldda, queue);
		}

		micSetDevice(1);

		/* == calling MAGMA with multiple MICs == */
		start = get_current_time();
		magma_cgetrf_mmic( num_mics, M, N, d_lA, ldda, new_ipiv, &info, queue);
		end = get_current_time();
		mic_perf = flops / GetTimerValue(start, end);
		if (info < 0) {
			printf("Argument %d of magma_cgetrf_mmic had an illegal value.\n", (int) -info);
			break;
		} else if (info != 0 ) {
			printf("magma_cgetrf_mmic returned info=%d.\n", (int) info);
			break;
		}

		/* == download the matrix from MICs == */
		for(int j = 0; j < N; j +=nb){
			k = (j / nb) % num_mics;
			micSetDevice(k + 1);
			nk = min(nb, N - j);
			magma_cgetmatrix( M, nk,
					d_lA[k]+j/(nb*num_mics)*nb*ldda, 0, ldda,
					h_R+j*lda,  0,                   lda, queue );
		}
		micSetDevice(1);

		/* =====================================================================
		   Check the factorization
		   =================================================================== */
		//error = get_LU_error(M, N, h_A, lda, h_R, new_ipiv);

		printf("%5d %5d  %6.2f         %6.2f         %e\n",
				(int) M, (int) N, cpu_perf, mic_perf, error);
		fflush(0);

		if (flag != 0)
			break;
	}

	/* Memory clean up */
	TESTING_FREE( ipiv );
	TESTING_FREE( h_A );
	TESTING_FREE( h_R );
	for(i = 0; i < num_mics0; i++){
		micSetDevice(i + 1);
		if(d_lA[i] != NULL)
			TESTING_FREE_DEV( d_lA[i] );
	}

	/* Shutdown */
	magma_queue_destroy(queue);
	magma_finalize();
}
